import os
import json
import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from PIL import Image
from pycocotools.coco import COCO
from torchvision import transforms
import torchvision.transforms as T

class CenterPad:
    def __init__(self, target_multiple=256):
       self.target_multiple = target_multiple
    
    def __call__(self, tensor):

        if len(tensor.shape) == 3:  
            _, height, width = tensor.shape
        elif len(tensor.shape) == 2: 
            height, width = tensor.shape
            
            tensor = tensor.unsqueeze(0)  
        else:
            raise ValueError("Input tensor must have 2 or 3 dimensions.")
        

        _, height, width = tensor.shape  # Assuming tensor has shape (C, H, W)
        
        pad_h = (self.target_multiple - height % self.target_multiple) % self.target_multiple
        pad_w = (self.target_multiple - width % self.target_multiple) % self.target_multiple
        
        pad_h_top = pad_h // 2
        pad_h_bottom = pad_h - pad_h_top
        pad_w_left = pad_w // 2
        pad_w_right = pad_w - pad_w_left
        
        padded_tensor = torch.nn.functional.pad(tensor, (pad_w_left, pad_w_right, pad_h_top, pad_h_bottom),
                                                value=0)
        return padded_tensor




class CocoSegmentationDataset(Dataset):
    def __init__(self, image_dir, coco_json , categories: list , transform = transforms.Compose([CenterPad(target_multiple=256)])):
        
        self.coco_json = coco_json
        self.coco = COCO(coco_json)
        self.image_dir = image_dir
        self.image_ids = self.coco.getImgIds()
        self.categories = categories
        self.transform = transform
  

    def __len__(self):
        return len(self.image_ids)



    def __getitem__(self, idx):
        
        img_id = self.image_ids[idx]
        img_info = self.coco.loadImgs([img_id])[0]
        img_path = os.path.join(self.image_dir, img_info['file_name'])
        image = Image.open(img_path).convert("RGB")
        masks = np.zeros((len(self.categories),image.height,image.width),dtype=np.uint8)
        
        for catIdx in range(len(self.categories)):

            mask = np.zeros((image.height,image.width),dtype=np.uint8)

            ann_ids = self.coco.getAnnIds(imgIds= [img_id], catIds = [self.categories[catIdx]])
            annotations = self.coco.loadAnns(ann_ids)
            for ann in annotations:
                ann_mask = self.coco.annToMask(ann)
                mask = np.maximum(mask,ann_mask)
           
            masks[catIdx,:,:] = mask
                                          
        image = torch.as_tensor(np.array(image), dtype = torch.float).permute(2,0,1) / 255
        masks = torch.as_tensor(np.array(masks), dtype=torch.float)

        image = self.transform(image)
        masks = self.transform(masks)
        
        return image, masks




from gwel.network import Segmenter
from gwel.network import Detector
import numpy as np
import cv2
from pycocotools import mask as mask_utils


class LociSegmenter(Segmenter):
    def __init__(self, detector: Detector,
                 bandwidth: int,
                 kernel_size: int):
        self.detector = detector
        self.bandwidth = bandwidth
        self.kernel_size = kernel_size
        self.channels = ['loci']
   
    def set_device(self, device : str):
        pass 

    def load_weights(self, weights : str = None):
        pass

    def inference(self, image : np.ndarray):
 
        #image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        H=image.shape[0]
        W=image.shape[1]
        objects = self.detector.inference(image)

        points = np.zeros((H, W), dtype=np.float32)
        uniform_kernel_points = points
        coords = np.array(objects)
        if len(coords) > 0:
            x = ((coords[:,0,0, 0, 0] + coords[:,0,2, 0, 0]) / 2).astype(int)
            y = ((coords[:,0,0, 0, 1] + coords[:,0,2, 0, 1]) / 2).astype(int)
        
            x = np.clip(x, 0, W - 1)
            y = np.clip(y, 0, H - 1)

            for i, j in zip(y, x):
                cv2.circle(uniform_kernel_points, (j, i), radius=self.bandwidth, color=1, thickness=-1)  
        
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (self.kernel_size*2+1, self.kernel_size*2+1))
        mask = cv2.morphologyEx(uniform_kernel_points, cv2.MORPH_CLOSE, kernel)
        rle = mask_utils.encode(np.asfortranarray(mask.astype(np.uint8)))
        output = {'loci': rle}
        return output
    
 
 



import torch
import numpy as np
import os
from torchvision import transforms
import cv2 
from gwel.UNET.unet import UNet 
from gwel.UNET.data_loader import CenterPad
from gwel.UNET.data_loader import CocoSegmentationDataset
from gwel.UNET.data_loader import DataLoader
import pandas as pd
from tqdm import tqdm

def load_model(model_path):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = UNet(in_channels=3, out_channels=3) 
    model.load_state_dict(torch.load(model_path, map_location=device,weights_only=True))
    model.to(device)
    model.eval()  
    return model, device

def load_image(image_path):
    image_bgr = cv2.imread(image_path)
    image = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
    image = torch.as_tensor(image,dtype=torch.float).permute(2,0,1) / 255
    return image


def display(image):
    image = np.uint8(image)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    cv2.imshow('image',image)

    # Wait for a key press to close the window
    cv2.waitKey(0)
    cv2.destroyAllWindows()


def run_inference(model, device, image, centerpad, threshold = 0.5,post_process=True):
    image_padded = centerpad(image)
    image_padded = image_padded.to(device)
    
    with torch.no_grad():  # Inference mode (no gradients)
       output = model(image_padded)
       output = torch.sigmoid(output)  # Apply sigmoid for binary segmentation
    
    output = output.cpu().numpy()
    
    masks =  np.array((output > threshold)*255,dtype=np.uint8)

    _, crop_height, crop_width = image.shape
    _, height, width = masks.shape
    
    left = (width - crop_width) // 2
    top = (height - crop_height) // 2
    right = (width + crop_width) // 2
    bottom = (height +crop_height) // 2 
    
    cropped_masks = masks[:,top:bottom, left:right]
    
    if post_process:
        cropped_masks = post_process_masks(cropped_masks)

    return cropped_masks

def post_process_masks(masks):
    mask1 = masks[0]  
    mask2 = masks[1]
    mask3 = masks[2]

    mask1 = (mask1 > 0).astype(np.uint8) * 255
    mask2 = (mask2 > 0).astype(np.uint8) * 255
    mask3 = (mask3 > 0).astype(np.uint8) * 255
    
   #num_labels, labels, stats, centroids = cv2.connectedComponentsWithStats(mask1, 8, cv2.CV_32S)
    #if np.max(mask1) > 0 :

    #	largest_component = 1 + np.argmax(stats[1:, cv2.CC_STAT_AREA])
    
    #	mask1 = (labels == largest_component).astype(np.uint8) * 255

    mask2 = mask2 * (mask1 // 255)
    mask3 = mask3 * (mask1 // 255)

    combined_masks = np.stack((mask1, mask2, mask3), axis=0)

    return combined_masks


def draw_contours(image,output,threshold=0.5):
    
    masks = np.split(output,output.shape[0])
    colours = [(255,255,0),(0,255,255)]
    image_with_contour=np.uint8(np.transpose(np.array(image)*255,(1,2,0))).copy()
    
    for idx in range(output.shape[0]):
        binary_image = np.squeeze(np.uint8((masks[idx]>threshold) * 255))
        contours,_  = cv2.findContours(binary_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        cv2.drawContours(image_with_contour, contours, -1, colours[idx] , 2) 
    return image_with_contour

def get_perimeter(mask,boundary=None):
    mask = np.squeeze(mask)
    mask = cv2.copyMakeBorder(mask, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=0)
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    perim = sum(cv2.arcLength(contour, True) for contour in contours) 
    
    if boundary is not None:
        boundary = np.squeeze(boundary)
        boundary = np.max(boundary)-boundary
        boundary = cv2.copyMakeBorder(boundary, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=255)
        contours, _ = cv2.findContours(boundary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        perim_boundary = sum(cv2.arcLength(contour, True) for contour in contours)
       
        mask_plus_boundary = mask | boundary
        contours, _ = cv2.findContours(mask_plus_boundary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        perim_neg = sum(cv2.arcLength(contour, True) for contour in contours) 

        perim = (perim + perim_neg - perim_boundary)/2  

    return perim


def anno_pred_comp(data_dir , data_json , model_path, csv_output,threshold):
    dataset = CocoSegmentationDataset(image_dir = data_dir, coco_json = data_json ,categories = [1,2,3])
    loader = DataLoader(dataset, batch_size=1)
    df = pd.DataFrame()
    model, device = load_model(model_path)
    for i, (inputs, targets) in enumerate(tqdm(loader)):
        with torch.no_grad():  
            masks = model(inputs)
            masks = torch.sigmoid(masks)
            masks =  np.array((masks > threshold)*255,dtype=np.uint8)
            targets = np.array((targets > threshold)*255,dtype=np.uint8)

            masks = masks.squeeze()
            targets = targets.squeeze()


            mask1,mask2,mask3 = masks[0], masks[1], masks[2]
            mask2 = mask2 * (mask1 // 255)
            mask3 = mask3 * (mask1 // 255)
            
            target1,target2,target3 = targets[0],targets[1],targets[2]
            target2 = target2 * (target1 // 255)
            mask3 = target3 * (target1 // 255)
            
            mask_area_1 = np.sum(mask1/255)
            mask_area_2 = np.sum(mask2/255)
            mask_area_3 = np.sum(mask3/255)
            
            mask_perim_1 = get_perimeter(mask1)
            mask_perim_2 = get_perimeter(mask2,mask1)
            mask_perim_3 = get_perimeter(mask3,mask1)
           
            target_area_1 = np.sum(target1/255)
            target_area_2 = np.sum(target2/255)
            target_area_3 = np.sum(target3/255)
            
            target_perim_1 = get_perimeter(target1)
            target_perim_2 = get_perimeter(target2, target1)
            target_perim_3 = get_perimeter(target3, target1)

            (height,width) = mask1.shape

            new_row = pd.DataFrame({'image_height':[height],
                                    'image_width':[width],
                                    'mask_area_1':[mask_area_1],
                                    'mask_area_2':[mask_area_2],
                                    'mask_area_3':[mask_area_3],
                                    'mask_perim_1':[mask_perim_1],
                                    'mask_perim_2':[mask_perim_2],
                                    'mask_perim_3':[mask_perim_3],
                                    'target_area_1':[target_area_1],
                                    'target_area_2':[target_area_2],
                                    'target_area_3':[target_area_3],
                                    'target_perim_1':[target_perim_1],
                                    'target_perim_2':[target_perim_2],
                                    'target_perim_3':[target_perim_3]
                                    
                                    })
            print(f'{target_area_3} and {mask_area_3} and {target_perim_3} and {mask_perim_3}')
            df = pd.concat([df,new_row],ignore_index=True)
        
    return df 


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='show model output from input image.')
    parser.add_argument('-i','--image_path', type=str, help='Path to image')
    parser.add_argument('-m','--model_path', type=str, help='Path to model')
    args = parser.parse_args()
    
    centerpad = CenterPad()

    image = load_image(args.image_path)
    
    model, device = load_model(args.model_path)

    output = run_inference(model, device, image, centerpad)

    img_with_contours = draw_contours(image,output)
    
    display(img_with_contours)







import argparse
import os
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
import torch.nn as nn
from tqdm import tqdm
from gwel.networks.UNET import UNet
from gwel.networks.unet.data_loader import CocoSegmentationDataset
import numpy as np
import ast 
import pandas as pd

def train(model, dataloader, criterion, optimizer, device, epoch, num_classes):
    model.train()
    epoch_loss = 0.0

    # Track TP, FP, FN per class
    iou_tp = torch.zeros(num_classes, device=device)
    iou_fp = torch.zeros(num_classes, device=device)
    iou_fn = torch.zeros(num_classes, device=device)

    for images, masks in dataloader:
        images = images.to(device)
        masks = masks.to(device).long()   # [N, H, W]

        #masks = masks.squeeze(0).long()
        
        optimizer.zero_grad()
        outputs = model(images)           # [N, C, H, W]

        # Loss
        loss = criterion(outputs, masks)
        loss.backward()
        optimizer.step()
        epoch_loss += loss.item()

        # Predictions
        preds = torch.argmax(outputs, dim=1)  # [N, H, W]

        # IoU statistics
        for c in range(num_classes):
            pred_c = (preds == c)
            mask_c = (masks == c)

            tp = (pred_c & mask_c).sum()
            fp = (pred_c & ~mask_c).sum()
            fn = (~pred_c & mask_c).sum()

            iou_tp[c] += tp
            iou_fp[c] += fp
            iou_fn[c] += fn

    avg_loss = epoch_loss / len(dataloader)

    # Compute IoU per class
    iou_per_class = []
    for c in range(num_classes):
        denom = (iou_tp[c] + iou_fp[c] + iou_fn[c]).item()
        iou = (iou_tp[c].item() / denom) if denom > 0 else float("nan")
        iou_per_class.append(iou)

    # Prepare metrics row
    metrics = {"epoch": epoch + 1, "loss": avg_loss}
    for c, iou in enumerate(iou_per_class):
        metrics[f"iou_class_{c}"] = iou

    return pd.DataFrame([metrics])


def validate(model, val_loader, criterion, device):
    model.eval()
    val_loss = 0.0
    with torch.no_grad():
        for inputs, targets in val_loader:
            inputs, targets = inputs.to(device), targets.to(device)
            outputs = model(inputs)
            loss = criterion(outputs, targets)
            val_loss += loss.item()

    print(f"Validation loss: {val_loss / len(val_loader):.4f}")


def main(args):

    os.makedirs(args.save_dir, exist_ok=True)

    categories = ast.literal_eval(args.categories)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


    # UNet with multi-class output
    model = UNet(in_channels=args.in_channels, out_channels=len(categories)+1).to(device)

    # Use CrossEntropyLoss for multi-class segmentation
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=args.lr)
    
    train_dataset = CocoSegmentationDataset(
        image_dir=args.train_images_dir, 
        coco_json=args.train_annotations,
        categories=categories
    )

    train_loader = DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True)

    # Training loop
    train_data = pd.DataFrame()

    start_epoch = 0

    # Load checkpoint if specified
    if args.resume_checkpoint is not None and os.path.exists(args.resume_checkpoint):
        checkpoint = torch.load(args.resume_checkpoint, map_location=device)
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        start_epoch = checkpoint.get('epoch', 0)  # resume from next epoch
        print(f"Resuming training from epoch {start_epoch + 1}")

        # Load previous metrics CSV if it exists
        if os.path.exists(args.train_metrics_path):
            train_data = pd.read_csv(args.train_metrics_path)

    for epoch in range(start_epoch, args.epochs):
        new_row = train(model, train_loader, criterion, optimizer, device, epoch, len(categories)+1)
        train_data = pd.concat([train_data, new_row], ignore_index=True)
        train_data.to_csv(args.train_metrics_path, index=False)


        # Save checkpoint
        if (epoch + 1) % args.save_interval == 0:
            checkpoint_path = os.path.join(args.save_dir, f"unet_epoch_{epoch + 1}.pth")
            torch.save({
                'epoch': epoch + 1,
                'model_state_dict': model.state_dict(),
                'optimizer_state_dict': optimizer.state_dict()
                    }, checkpoint_path)
            print(f"Saved checkpoint: {checkpoint_path}")




if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Train a U-Net model")

    parser.add_argument('-td','--train_images_dir', type=str, required=True, help="Path to directory of training images")
    parser.add_argument('-ta','--train_annotations', type=str, required=True, help="Path to training annotations")
    parser.add_argument('-c','--categories', type=str, required=True, help="Integer tag for annotation categories")
    parser.add_argument('-s','--save_dir', type=str, required=True, help="Directory to save model checkpoints")

    parser.add_argument('--epochs', type=int, default=10, help="Number of epochs")
    parser.add_argument('--batch_size', type=int, default=16, help="Batch size")
    parser.add_argument('--lr', type=float, default=1e-4, help="Learning rate")
    parser.add_argument('--in_channels', type=int, default=1, help="Number of input channels (1 for grayscale images)")
    parser.add_argument('--out_channels', type=int, default=1, help="Number of output channels (1 for binary segmentation)")
    parser.add_argument('--save_interval', type=int, default=1, help="How often to save model checkpoints (in epochs)")
    parser.add_argument('--val_interval', type=int, default=1, help="How often to validate the model (in epochs)")
    parser.add_argument('-vd','--val_images_dir', type=str,default=None, help="Path to directory of validation images")
    parser.add_argument('-va','--val_annotations', type=str, default=None, help="Path to validation annotations") 
    parser.add_argument('--train_metrics_path',type=str,default="train-data.csv",help="Path to output train metrics csv")
    parser.add_argument('--resume_checkpoint', type=str, default=None, help="Path to a checkpoint .pth to resume training")
    args = parser.parse_args()

    # Check that paths exist
    assert os.path.exists(args.train_images_dir), f"Train images path '{args.train_images_dir}' not found."
    assert os.path.exists(args.train_annotations), f"Train masks path '{args.train_annotations}' not found."
    os.makedirs(args.save_dir, exist_ok=True)

    # Start training
    main(args)




import numpy as np
import cv2
import onnxruntime as ort
import warnings

from gwel.network import Detector

warnings.filterwarnings("ignore")


def bbox_to_polygon(bboxes):
    polygons = []
    for bbox in bboxes:
        x1, y1, x2, y2 = bbox
        polygon = [[x1, y1], [x1, y2], [x2, y2], [x2, y1]]
        polygons.append([np.array(polygon, dtype=np.int32).reshape(-1, 1, 2)])
    return polygons


def non_max_suppression(boxes, scores, iou_threshold=0.45):
    # boxes: [N,4] numpy array
    # scores: [N] numpy array
    idxs = scores.argsort()[::-1]
    keep = []

    while len(idxs) > 0:
        i = idxs[0]
        keep.append(i)
        if len(idxs) == 1:
            break
        ious = compute_iou(boxes[i], boxes[idxs[1:]])
        idxs = idxs[1:][ious < iou_threshold]

    return keep


def compute_iou(box, boxes):
    x1 = np.maximum(box[0], boxes[:, 0])
    y1 = np.maximum(box[1], boxes[:, 1])
    x2 = np.minimum(box[2], boxes[:, 2])
    y2 = np.minimum(box[3], boxes[:, 3])

    inter_area = np.maximum(0, x2 - x1) * np.maximum(0, y2 - y1)
    box_area = (box[2] - box[0]) * (box[3] - box[1])
    boxes_area = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])
    union_area = box_area + boxes_area - inter_area
    return inter_area / (union_area + 1e-6)


class YOLOv8(Detector):
    def __init__(self, weights: str, device: str = "cpu", patch_size: tuple = None, threshold: float = 0.5):
        self.device = device
        self.patch_size = patch_size
        self.threshold = threshold
        self.weights = weights
        self.session = ort.InferenceSession(weights, providers=["CPUExecutionProvider"])

    def inference(self, image: np.ndarray):
        if self.patch_size:
            return self.inference_with_patches(self.patch_size, image)

        input_tensor = self.preprocess(image)
        outputs = self.session.run(None, {"images": input_tensor})
        detections = self.postprocess(outputs, image.shape[:2])
        return bbox_to_polygon(detections)

    def inference_with_patches(self, patch_size: tuple, image: np.ndarray):
        h, w = image.shape[:2]
        patch_h, patch_w = patch_size
        pad_h = (patch_h - h % patch_h) % patch_h
        pad_w = (patch_w - w % patch_w) % patch_w
        padded_image = cv2.copyMakeBorder(image, 0, pad_h, 0, pad_w, cv2.BORDER_CONSTANT, value=(0, 0, 0))

        detections = []
        for i in range(0, padded_image.shape[0], patch_h):
            for j in range(0, padded_image.shape[1], patch_w):
                patch = padded_image[i:i + patch_h, j:j + patch_w]
                input_tensor = self.preprocess(patch)
                outputs = self.session.run(None, {"images": input_tensor})
                boxes = self.postprocess(outputs, patch.shape[:2])
                # adjust boxes to original coordinates
                for bbox in boxes:
                    x1, y1, x2, y2 = bbox
                    x1 += j
                    x2 += j
                    y1 += i
                    y2 += i
                    x1, y1 = max(0, x1), max(0, y1)
                    x2, y2 = min(w, x2), min(h, y2)
                    detections.append([x1, y1, x2, y2])

        return bbox_to_polygon(detections)

    def preprocess(self, image: np.ndarray):
        img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, (640, 640))
        img = img.astype(np.float32) / 255.0
        img = np.transpose(img, (2, 0, 1))
        return np.expand_dims(img, axis=0)

    def postprocess(self, outputs, original_shape):
        # ONNX YOLOv8 raw output: [batch, num_boxes, 85]
        # 85 = [x, y, w, h, conf, class_probs...]
        predictions = outputs[0][0]  # take first batch
        boxes = predictions[:, :4]
        scores = predictions[:, 4] * predictions[:, 5:].max(axis=1)
        mask = scores > self.threshold
        boxes = boxes[mask]
        scores = scores[mask]
        if len(boxes) == 0:
            return []
        keep = non_max_suppression(boxes, scores)
        # Convert xywh to xyxy
        xyxy_boxes = []
        for i in keep:
            x, y, w, h = boxes[i]
            xyxy_boxes.append([x, y, x + w, y + h])
        return xyxy_boxes




import torch
import torch.nn as nn
import torch.nn.functional as F

class UNet(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(UNet, self).__init__()

        # Contracting path (Encoder)
        self.encoder1 = self.conv_block(in_channels, 64)
        self.encoder2 = self.conv_block(64, 128)
        self.encoder3 = self.conv_block(128, 256)
        self.encoder4 = self.conv_block(256, 512)

        # Expanding path (Decoder)
        self.decoder4 = self.upconv_block(512, 256)
        self.decoder3 = self.upconv_block(256, 128)
        self.decoder2 = self.upconv_block(128, 64)
        self.decoder1 = self.out_layer(64, out_channels)


    def conv_block(self, in_channels, out_channels):
        return nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=4,stride=2,padding=1)
        )

    def upconv_block(self, in_channels, out_channels):
        return nn.Sequential(
            nn.ConvTranspose2d(in_channels, out_channels, kernel_size=4, stride=2,padding=1),
            nn.ReLU(inplace=True),
	    nn.Conv2d(out_channels,out_channels,kernel_size=3,padding=1),
            nn.ReLU(inplace=True)
        )
    
    def out_layer(self,in_channels,out_channels):
        return nn.Sequential(
	    nn.ConvTranspose2d(in_channels,out_channels,kernel_size=4,stride = 2,padding=1),
	    nn.Conv2d(out_channels,out_channels,kernel_size=3,padding=1)
        )

    def forward(self, x):
        # Encoder
        e1 = self.encoder1(x)
        e2 = self.encoder2(e1)
        e3 = self.encoder3(e2)
        e4 = self.encoder4(e3)

        # Decoder
        d3 = self.decoder4(e4)
        d2 = self.decoder3(d3 + e3)  
        d1 = self.decoder2(d2 + e2)   
        out = self.decoder1(d1 + e1)   
        return out



import torch
import torchvision
from torchvision import transforms
from PIL import Image

# Function to preprocess the image
def process_image(image_path):
    image = Image.open(image_path).convert("RGB")

    transform = transforms.Compose([
        transforms.ToTensor(),  # Convert the image to a tensor
    ])

    image_tensor = transform(image)
    return image_tensor

# Function to extract bounding boxes from the predictions
def get_bboxes(model, image_path, threshold=0.5, device="cpu"):
    # Preprocess the image
    image_tensor = process_image(image_path)

    # Add batch dimension (model expects a batch of images)
    image_tensor = image_tensor.unsqueeze(0).to(device)

    # Run inference
    with torch.no_grad():
        predictions = model(image_tensor)

    # Get predicted bounding boxes and scores
    pred_boxes = predictions[0]['boxes'].cpu().numpy()  # Bounding boxes
    pred_scores = predictions[0]['scores'].cpu().numpy()  # Prediction scores

    # Filter bounding boxes by confidence score
    filtered_boxes = []
    for box, score in zip(pred_boxes, pred_scores):
        if score > threshold:  # Only include boxes with score above the threshold
            filtered_boxes.append(box)

    return filtered_boxes



import os
from tqdm import tqdm
import torch
import torchvision
from torchvision import datasets, transforms, models
from torch.utils.data import DataLoader
from torchvision.models.detection import FasterRCNN
from torchvision.models.detection import fasterrcnn_resnet50_fpn
from pycocotools.coco import COCO
import numpy as np
from torch.utils.data import Dataset
import json
from PIL import Image
from gwel.FRCNN.dataset import CocoDataset

import torch
import numpy as np

def iou(box1, box2):
    """Calculate the Intersection over Union (IoU) of two bounding boxes."""
    x1, y1, x2, y2 = box1
    x1_gt, y1_gt, x2_gt, y2_gt = box2

    # Compute the area of intersection
    inter_x1 = max(x1, x1_gt)
    inter_y1 = max(y1, y1_gt)
    inter_x2 = min(x2, x2_gt)
    inter_y2 = min(y2, y2_gt)
    
    inter_area = max(0, inter_x2 - inter_x1) * max(0, inter_y2 - inter_y1)
    
    # Compute the area of both boxes
    box1_area = (x2 - x1) * (y2 - y1)
    box2_area = (x2_gt - x1_gt) * (y2_gt - y1_gt)
    
    # Compute IoU
    union_area = box1_area + box2_area - inter_area
    return inter_area / union_area if union_area > 0 else 0

def compute_precision_recall(pred_boxes, gt_boxes, pred_labels, gt_labels, iou_threshold=0.5):
    """
    Compute Precision and Recall based on bounding box predictions and ground truth boxes.

    Arguments:
    - pred_boxes: List of tensors of predicted bounding boxes for each image [N, 4]
    - gt_boxes: List of tensors of ground truth bounding boxes for each image [M, 4]
    - pred_labels: List of tensors of predicted labels for each image [N]
    - gt_labels: List of tensors of ground truth labels for each image [M]
    - iou_threshold: IoU threshold to classify true positives

    Returns:
    - precision: Precision score
    - recall: Recall score
    """
    tp = 0
    fp = 0
    fn = 0

    for image_pred_boxes, image_gt_boxes, image_pred_labels, image_gt_labels in zip(pred_boxes, gt_boxes, pred_labels, gt_labels):
        # Match predicted boxes to ground truth boxes
        matched_gt_boxes = set()  # Keep track of matched ground truth boxes
        
        for i, pred_box in enumerate(image_pred_boxes):
            pred_label = image_pred_labels[i]
            
            best_iou = 0
            best_gt_idx = -1
            
            # Find the best matching ground truth box
            for j, gt_box in enumerate(image_gt_boxes):
                gt_label = image_gt_labels[j]
                
                if pred_label == gt_label:  # Only match boxes of the same class
                    iou_score = iou(pred_box, gt_box)
                    if iou_score > best_iou:
                        best_iou = iou_score
                        best_gt_idx = j

            # If the IoU is above the threshold, it's a true positive
            if best_iou >= iou_threshold:
                tp += 1  # True positive: valid match
                matched_gt_boxes.add(best_gt_idx)  # Mark this ground truth box as matched
            else:
                fp += 1  # False positive: no valid match
        
        # False negatives are the unmatched ground truth boxes
        fn += len(image_gt_boxes) - len(matched_gt_boxes)

    # Compute precision and recall
    precision = tp / (tp + fp) if (tp + fp) > 0 else 0
    recall = tp / (tp + fn) if (tp + fn) > 0 else 0
    
    return precision, recall




def FRCNNtrain(images_dir, annotations_file, batch_size=1, num_epochs = 10):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Load pre-trained Faster R-CNN model with ResNet-50 backbone and FPN
    model = fasterrcnn_resnet50_fpn(pretrained=True)

    # Update the model's classifier to fit the number of classes in your dataset
    num_classes = 2  # Change this according to your dataset (including background as class 0)
    in_features = model.roi_heads.box_predictor.cls_score.in_features
    model.roi_heads.box_predictor = torchvision.models.detection.faster_rcnn.FastRCNNPredictor(in_features, num_classes)

    model.to(device)

    params = [p for p in model.parameters() if p.requires_grad]
    optimizer = torch.optim.Adam(params, lr=0.0001)

    lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=3, gamma=0.1)


    ## dataloader change this depending on the dataset
    img_dir = images_dir
    ann_file = annotations_file

    transform = transforms.Compose([
        transforms.ToTensor(),  # Converts the image to a tensor
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])  # Normalizes the image
    ])

    dataset = CocoDataset(img_dir=img_dir, ann_file=ann_file, transform=transform)
    data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True, collate_fn=lambda x: tuple(zip(*x)))



    for epoch in range(num_epochs):
        model.train()
        
        total_loss = 0.0
        total_precision = 0.0
        total_recall = 0.0
        total_samples = 0

        # Wrap your DataLoader in tqdm to show a progress bar
        for images, targets in tqdm(data_loader, desc=f"Training Epoch {epoch + 1}/{num_epochs}", total=len(data_loader)):
            images = [image.to(device) for image in images]
            targets = [{k: v.to(device) for k, v in t.items()} for t in targets]
            
            optimizer.zero_grad()
            
            # Forward pass
            loss_dict = model(images, targets)
            
            # Compute total loss
            losses = sum(loss for loss in loss_dict.values())
            total_loss += losses.item()

            # Backward pass and optimization
            losses.backward()
            optimizer.step()

            # Get predictions
            with torch.no_grad():
                model.eval()
                predictions = model(images)
                model.train()

            # Collect predicted boxes and labels
            pred_boxes = [prediction['boxes'] for prediction in predictions]
            pred_labels = [prediction['labels'] for prediction in predictions]

            # Collect ground truth boxes and labels
            gt_boxes = [target['boxes'] for target in targets]
            gt_labels = [target['labels'] for target in targets]

            # Compute Precision and Recall
            precision, recall = compute_precision_recall(pred_boxes, gt_boxes, pred_labels, gt_labels, iou_threshold=0.5)
            
            total_precision += precision
            total_recall += recall
            total_samples += 1
            
        # Average Precision and Recall for the epoch
        avg_loss = total_loss / len(data_loader)
        avg_precision = total_precision / total_samples
        avg_recall = total_recall / total_samples

        # Print epoch metrics
        print(f"Epoch [{epoch + 1}/{num_epochs}] - Loss: {avg_loss:.4f}, Precision: {avg_precision:.4f}, Recall: {avg_recall:.4f}")
        
    return model 





import os 
import torch
from PIL import Image
from pycocotools.coco import COCO
from torch.utils.data import Dataset
import json

class CocoDataset(Dataset):
    def __init__(self, img_dir, ann_file, transform=None):
        self.img_dir = img_dir
        self.transform = transform
        
        # Load COCO annotations
        with open(ann_file, 'r') as f:
            self.coco_data = json.load(f)
        
        self.coco = COCO(ann_file)
        self.img_ids = list(self.coco.imgs.keys())
        
    def __len__(self):
        return len(self.img_ids)

    def __getitem__(self, idx):
        img_id = self.img_ids[idx]
        img_info = self.coco.imgs[img_id]
        img_path = os.path.join(self.img_dir, img_info['file_name'])
        
        # Load image
        image = Image.open(img_path).convert("RGB")
        
        # Get annotations for the image
        ann_ids = self.coco.getAnnIds(imgIds=img_id)
        anns = self.coco.loadAnns(ann_ids)
        
        boxes = []
        labels = []
        for ann in anns:
            x, y, w, h = ann['bbox']
            boxes.append([x, y, x + w, y + h])  # COCO format: [x, y, x+w, y+h]
            # If your annotations are just the background (0), change them to 1 for object class
            labels.append(1)  # Object class label set to 1
        
        boxes = torch.tensor(boxes, dtype=torch.float32)
        labels = torch.tensor(labels, dtype=torch.int64)
        
        target = {'boxes': boxes, 'labels': labels}
        
        if self.transform:
            image = self.transform(image)
        
        return image, target



from abc import ABC, abstractmethod
import numpy as np


class Network(ABC):
    @abstractmethod
    def load_weights(self):
        pass

    @abstractmethod
    def inference(self, image : np.ndarray):
        pass

    @abstractmethod
    def set_device(self, device : str):
        pass


class Detector(Network, ABC):
    pass 
    
    #def inference_with_patches(self, image : np.ndarray , patch_size : tuple[int,int]):
        #pass
    

    #def validate_inference():
        #pass

class Segmenter(Network,ABC):
    pass
   

    #def validate_inference():




import torch
import numpy as np
import matplotlib.pyplot as plt
from unet import UNet  # Import your UNet model
from data_loader import CustomDataset  # Import your custom dataset class
from torch.utils.data import DataLoader
from torchvision import transforms
from sklearn.metrics import jaccard_score  # For IoU score

# Hyperparameters
batch_size = 8

# Prepare the dataset and DataLoader for validation
image_dir = 'path/to/val_images'
label_dir = 'path/to/val_labels'
transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
])

val_dataset = CustomDataset(image_dir, label_dir, transform)
val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)

# Load the trained model
model = UNet(in_channels=3, out_channels=1).cuda()
model.load_state_dict(torch.load('unet_epoch_50.pth'))  # Load the model checkpoint
model.eval()  # Set the model to evaluation mode

# Evaluation loop
all_preds = []
all_labels = []

with torch.no_grad():  # No need to compute gradients for evaluation
    for images, labels in val_loader:
        images, labels = images.cuda(), labels.cuda()
        
        # Forward pass
        outputs = model(images)
        
        # Convert logits to binary predictions
        preds = torch.sigmoid(outputs) > 0.5
        
        all_preds.append(preds.cpu().numpy())
        all_labels.append(labels.cpu().numpy())

# Convert lists to numpy arrays
all_preds = np.concatenate(all_preds, axis=0)
all_labels = np.concatenate(all_labels, axis=0)

# Calculate the Jaccard Index (IoU)
iou = jaccard_score(all_labels.flatten(), all_preds.flatten(), average='binary')
print(f"Validation IoU: {iou:.4f}")

# Visualize some of the results
def visualize(image, label, pred):
    fig, axes = plt.subplots(1, 3, figsize=(12, 4))
    axes[0].imshow(image.permute(1, 2, 0).cpu().numpy())
    axes[0].set_title("Image")
    axes[1].imshow(label.squeeze(), cmap='gray')
    axes[1].set_title("Ground Truth")
    axes[2].imshow(pred.squeeze(), cmap='gray')
    axes[2].set_title("Prediction")
    plt.show()

# Visualizing a few random examples
for i in range(3):
    idx = np.random.randint(0, len(val_dataset))
    image, label = val_dataset[idx]
    pred = torch.sigmoid(model(image.unsqueeze(0).cuda())).squeeze().cpu().detach() > 0.5
    visualize(image, label, pred)




import os
import shutil
import cv2
import glob
from PIL import Image
from tqdm import tqdm
import random
import json
import numpy as np
import pandas as pd
from gwel.network import Detector, Segmenter 
import sys
import os
from colorama import Fore, Style, init
import datetime
from pycocotools import mask as maskUtils

# Initialize colorama
init(autoreset=True)

hidden_file_name = ".gwel"
Image.MAX_IMAGE_PIXELS = None

def scale_contours(contours, scale_x, scale_y):
    scaled_contours = []
    for contour in contours:
        if np.issubdtype(contour.dtype, np.integer):
            contour = contour.astype(np.float32)
            contour = contour + 0.5
        if len(contour.shape) > 2:
            contour = contour[0, :] if contour.shape[0] == 1 else np.squeeze(contour)      
        contour[:, 0] *= scale_x 
        contour[:, 1] *= scale_y
        if scale_x * scale_y > 1:
            contour = np.round(contour).astype(np.int32)
        scaled_contours.append(contour)
    return scaled_contours

class ImageDataset:
    def __init__(self, directory: str, check : bool = False):
       

        self.directory = directory
        self.images = self._get_image_list(self.directory)
        if len(self.images) == 0:
            raise ValueError(f"No images found in directory: {directory}")
        self.image_sizes = self._get_image_sizes(self.directory)
        self.resized_image_sizes = {} 
        self.resized_directory = ''
        self.patches = []
        self.patch_size = None
        self.object_detections = {}
        self.masks = {}
        self.flagged = []
        self.annotations = False
        self.factors = {}
        self.warp = {}
        
        os.makedirs(os.path.join(self.directory, hidden_file_name), exist_ok=True)

        # Check for problematic images and filter them out
        if check:
            self.check_images()

    def _get_image_list(self, directory : str):
        return [f for f in os.listdir(directory) if f.lower().endswith(('.jpg', '.png', '.jpeg', '.bmp', '.tiff','.tif', '.JPG'))]
    
    def _get_image_paths(self,directory : str):
        image_extensions = ['*.jpg', '*.jpeg', '*.png', '*.gif', '*.bmp', '*.tiff','*.JPG','*.tif']
        image_paths = [] 
        for ext in image_extensions:
            image_paths.extend(glob.glob(os.path.join(directory, ext)))
        return image_paths
    

    def _get_image_sizes(self, directory : str):  
        image_sizes = {}
        for image_name in tqdm(self.images, desc="Retriving image sizes", unit="image"):
            img = Image.open(os.path.join(directory ,image_name))
            if hasattr(img, '_getexif'):
                exif = img._getexif()
                width, height = (img.size[1], img.size[0]) if exif and exif.get(274) in [6, 8] else img.size
            else:
                width, height = (img.size[0], img.size[1])    
            image_sizes[image_name] = [height, width]
        return image_sizes

    def _check_image(self, image_name):
        """Check if a single image can be read."""
        img_path = os.path.join(self.directory, image_name)
       
        return True if cv2.imread(img_path) is not None else False

    def check_images(self):
        """Check if images can be read and report any problematic files."""
        problematic_images = []

        for image in tqdm(self.images, desc = "Checking images", unit = "image"):
            check = self._check_image(image)
            if not check: 
                problematic_images.append(image)

        self.images = [img for img in self.images if img not in problematic_images]

        if problematic_images:
            print("Warning: Unable to read the following images:")
            print("\n".join(f"- {img}" for img in problematic_images))

    def _print_status(self, condition, message):
        """Helper function to print the status with appropriate colors."""
        if condition:
            print(f"{Fore.GREEN}{Style.BRIGHT}{message} ✓{Style.RESET_ALL}")
        else:
            print(f"{Fore.RED}{Style.BRIGHT}{message} X{Style.RESET_ALL}")

    def status(self):
        print("=" * 25)
        print(f"{Style.BRIGHT}{Fore.CYAN}--- STATUS ---{Style.RESET_ALL}")
        self._print_status(self.directory, "Directory")
        self._print_status(self.images, "Images list")
        self._print_status(self.image_sizes, "Image Sizes")
        self._print_status(self.factors, f"Factors ({len(self.factors)})")
        self._print_status(self.resized_image_sizes, "Resized Images")
        self._print_status(self.object_detections, "Object Detections")
        self._print_status(self.masks, "Segmentation")
        print("=" * 25)  

    def factor(self, name: str, values : list):
        self.factors[name]=values

    def sort(self, factors = []):
        self.images = self._get_image_list(self.directory)
        N = len(self.images)
        order = np.arange(N)
        if factors:
            multiplier = 1
            for factor in factors:
                values = self.factors[factor]
                value_set = sorted(list(set(values)), reverse = True) # sort so alpha comes before numeric in string values
                multiplier = N * multiplier
                if isinstance(list(value_set)[0], str):
                    values = [value_set.index(value) for value in values]
                values = np.array(values)
                order = order + multiplier * values
        
        order = list(order)
        self.images = [ x for _,x in sorted(zip(order,self.images))]

    def tree(self, output_dir, factors):
        df = pd.DataFrame(self.factors)
        df["image"] =  self._get_image_list(self.directory)
        for factor_values, df_group in tqdm(df.groupby(factors)):
            path_parts = map(str, factor_values) 
            dir_path = os.path.join(output_dir, *path_parts)
            os.makedirs(dir_path, exist_ok=True)
            images = df_group['image']
            for image in images:
                image_path=os.path.join(self.directory,image)
                new_image_path= os.path.join(dir_path,image)
                shutil.copy(image_path, new_image_path)


    def unflag(self):
        self.flagged = []


    def flip(self, flipped_directory=''):
        response = input("Are you sure you want to flip the currently flagged images? "
                     "Saved annotations may not be flipped if not loaded and saved after flip (y/n): ").strip().lower()

        if response != 'y':
            print("Aborted flipping.")
            return 

        if flipped_directory:
            os.makedirs(os.path.join(self.directory,flipped_directory), exist_ok=True)
        for image in self.flagged:
            image_path = os.path.join(self.directory, image)
            img = cv2.imread(image_path)
            img_flipped =cv2.rotate(img,cv2.ROTATE_180)
            image_flipped_path = os.path.join(self.directory,flipped_directory,image)
            cv2.imwrite(image_flipped_path,img_flipped)
            if self.resized_directory:
                if os.path.exists(os.path.join(self.resized_images_directory,image)): 
                    image_path = os.path.join(self.resized_images_directory,image)
                    img = cv2.imread(image_path)
                    img_flipped =cv2.rotate(img,cv2.ROTATE_180)
                    image_flipped_path = os.path.join(self.directory,self.resized_images_directory,image)
                    if not flipped_directory:
                        cv2.imwrite(image_flipped_path,img_flipped)
        self.unflag()

    
    def filter(self,factors = {}):
        result = []
        images_list = self._get_image_list(self.directory)
        for factor, values in factors.items():
            if factor in self.factors:  
                indices = [i for i, val in enumerate(self.factors[factor]) if val in values]
                result.extend(images_list[i] for i in indices)
            else:
                print(f'Unknown factor: {factor}.')
                return
        return result


    def sample(self, directory : str, N : int = 0, flagged : bool = True, factors : list = {}, resized = False):

        images_list = self._get_image_list(self.directory)
        
        if os.path.exists(directory):
            choice = input("This directory already exists. Do you want to continue and use it anyway? (y/n): ").strip().lower()
            if choice != 'y':
                print("Sampling cancelled.")
                return
        else:
            os.makedirs(directory)

        if factors:
            images_list = self.filter(factors)
            if N == 0:
                N = len(images_list)
        
      
        if flagged:
            N_flagged = len(self.flagged)
            if N < N_flagged:
                print(f'Sample size is less then the number of flagged images. Do you need flagged = False, or set N = {len(self.flagged)}?')
                return
            images = random.sample(images_list,N - N_flagged)
            images.extend(self.flagged)
        else:
            images = random.sample(images_list, N)
        
        
        if N ==0:
            print('N must be a postive integer')
            return
        if resized:
            directory_original = os.path.join(self.resized_images_directory) 
        else:
            directory_original = self.directory

        for image in images:
            shutil.copy(os.path.join(directory_original,image),os.path.join(directory,image))

    def resize(self, resized_directory: str = os.path.join(hidden_file_name,"resized"), max_size = 800):
        os.makedirs(os.path.join(self.directory,resized_directory), exist_ok=True)
        os.makedirs(os.path.join(self.directory,resized_directory,"images"), exist_ok=True)
        self.resized_directory = os.path.join(self.directory,resized_directory)
        self.resized_images_directory = os.path.join(self.directory,resized_directory,"images")

        for image_name in tqdm(self.images, desc="Resizing images", unit="image"):
            resized_img_path = os.path.join(self.directory, resized_directory,"images", image_name)
            if os.path.exists(resized_img_path):
                continue
            img = cv2.imread(os.path.join(self.directory, image_name))
            ratio = min(max_size / img.shape[1], max_size / img.shape[0], 1)
            resized_img = cv2.resize(img, (int(img.shape[1] * ratio), int(img.shape[0] * ratio)), interpolation=cv2.INTER_AREA)
            cv2.imwrite(resized_img_path, resized_img)

        self.resized_image_sizes = self._get_image_sizes(os.path.join(self.directory, resized_directory,"images"))


    def patch_images(self,patch_size, patch_directory: str = os.path.join(hidden_file_name,"patches")):
        os.makedirs(os.path.join(self.directory,patch_directory,"images"), exist_ok=True)
        self.patch_size = patch_size
        self.output_directories["patches"] = patch_directory
        self.output_directories["patch_images"] = os.path.join(patch_directory,"images")
        self.patch_size = patch_size
        for image_name in tqdm(self.images, desc="Patching images", unit="image"):
            img = cv2.imread(os.path.join(self.directory, image_name))

            img_size = self.image_sizes[image_name]
            pad_H = (patch_size - img_size[0] % patch_size) % patch_size
            pad_W = (patch_size - img_size[1] % patch_size) % patch_size
            padded_img = np.pad(img, ((0, pad_H), (0, pad_W), (0, 0)), mode='constant', constant_values=0)
            patches = patchify(padded_img, [patch_size, patch_size, img.shape[2]], step=patch_size)
            patch_count = 0
            for i, patch_row in enumerate(patches):
               for j, patch in enumerate(patch_row):
                    patch_filename = os.path.join(self.directory,patch_directory,"images", f"{image_name}_patch_{i}_{j}.png")
                    print(patch_filename)
                    cv2.imwrite(patch_filename, patch[0])
    
    def sample_patches(self, N : int , dir : str, flagged : bool = True):
        if os.path.exists(dir):
            print('This directory already exists, perhaps try a different one?')
            return
        os.makedirs(dir)
        patch_dir = self.output_directories["patch_images"] 
        patches_list = os.listdir(patch_dir)
        images = random.sample(patches_list, N)
        for image in images:
            shutil.copy(os.path.join(self.directory,patch_dir,image),os.path.join(dir,image))


    def detect(self, detector : Detector = None, use_saved : bool = True, annotations_path : str = None, write : bool = True, threshold = 0.9):
        resized_dir = self.resized_directory
        if use_saved == True:
            pre_saved = self.load_object_detections(annotations_path)
            if pre_saved:
                return
            else:
                print("Pre-saved annotations not found.")
            
            
        if not resized_dir:
            print('Try resizing the images first')
            return
        else:
            resized_dir = self.resized_images_directory
 
       
        if not detector:
            raise ValueError("No Detector object provided.")

        print("Detecting Objects...")
        
        
        detector.threshold = threshold 

        for image_name in tqdm(self.images, desc="Processing images", unit="image"):
            
            self.object_detections[image_name]={"image_size": None, "polygons": []}
            img = cv2.imread(os.path.join(resized_dir, image_name))
            detected_instances = detector.inference(img) 
            height, width, _ = img.shape

            self.object_detections[image_name]["image_size"] = (height, width)
            for instance in detected_instances:
                self.object_detections[image_name]["polygons"].append(instance)               
        if write: 
            self.write_object_detections()
            #self.write_object_detections(resized=True)




    def load_object_detections(self, annotations_file :str , write : bool = False, add = False):
        
        if not annotations_file:
            subdir = self.resized_directory if getattr(self, "resized_directory", "") else hidden_file_name
            annotations_file = os.path.join(self.directory, subdir, "detections_coco.json")
        if not os.path.exists(annotations_file):
            return False

        with open(annotations_file, 'r') as f:
            coco_data = json.load(f)
        
        image_info = {image["id"]: (image["file_name"], image["width"], image["height"]) for image in coco_data["images"]}
        
        if not add:
            for image_name in self.images: self.object_detections[image_name] = {"image_size": None, "polygons": []}
        else:
            for image in coco_data["images"]:  self.object_detections[image["file_name"]] = {"image_size": None, "polygons": []}

         
        for annotation in tqdm(coco_data["annotations"], desc="Loading annotations", unit="annotation"):
            image_id = annotation["image_id"]
            image_name, width, height = image_info[image_id]
            segmentation = annotation["segmentation"]
            contours = []
            for segment in segmentation:
                segment = np.array(segment,dtype=np.int32).reshape(-1, 2)
                contours.append(segment)
            if image_name in self.images:
                self.object_detections[image_name]["polygons"].append(contours) 
                self.object_detections[image_name]["image_size"] = (height, width)

        if write:
            #self.write_object_detections()
            self.write_object_detections(resized = True)
        
        return True
  
    def write_object_detections(self, output_file : str = None, resized : bool = False, overwrite : bool = True, name : str = 'leaf', supercategory : str = 'plant'):

        coco_data = {
                "licenses": [{"name":"","id":0,"url":""}],
                "info": {"contributor":"","date_created":"","description":"","url":"","version":"","year":""},
                "images": [],
                "annotations": [],
                "categories": [{"id": 1, "name": name, "supercategory": supercategory}]
            }
        
        annotation_id = 1
        image_id = 1

        for image_name in tqdm(self.images, desc="Writing COCO annotations", unit="image"):
            
            width, height = self.resized_image_sizes[image_name] if resized else self.image_sizes[image_name]

            coco_data["images"].append({
                "id": image_id,
                "file_name": image_name,
                "height": width,
                "width": height
            })
            
            detections  = self.object_detections.get(image_name, []) 
            for instance in detections["polygons"]: 
                segmentation = []
                all_x, all_y = [], []

                for polygon in instance:
                    #polygon = np.squeeze(polygon)
                    #if isinstance(polygon[0], (list, tuple, np.ndarray)):
                    #segment = [int(c) for p in polygon for c in p]
                    polygon = polygon.reshape(-1, 2)  # shape becomes (4,2)
                    xs, ys = zip(*polygon)             # now works
                    segment = polygon.flatten().astype(int).tolist()
                    #else:
                     #   print(annotation_id)
                      #  print(len(polygon))
                       # segment = [int(v) for v in polygon]
                        #xs, ys = segment[::2], segment[1::2]
                    segmentation.extend(segment)
                    all_x.extend(xs)
                    all_y.extend(ys)

                if len(segmentation) < 6:
                    continue

                x_min, y_min = int(min(all_x)), int(min(all_y))
                width, height = int(max(all_x)-x_min), int(max(all_y)-y_min)

                coco_data["annotations"].append({
                    "id": annotation_id,
                    "image_id": image_id,
                    "category_id": 1,
                    "segmentation": [segmentation],
                    "bbox": [x_min, y_min, width, height],
                    "area": width*height,
                    "iscrowd": 0
                })
                annotation_id += 1
  
            image_id += 1

        
        
        if not output_file: 
            output_file = os.path.join(self.directory,hidden_file_name,"detections_coco.json")
            resized_dir =  self.resized_directory
            #if resized and resized_dir:
                #output_file = os.path.join(resized_dir ,"detections_coco.json")
        
        if not output_file:
            print('Could not write with no output file defined. For default write locations ensure overwrite = True')
            return

        with open(os.path.join(output_file), 'w') as f:
            json.dump(coco_data, f, indent=4)
	
        
    def crop(self, output_directory: str, object_name : str):
        if os.path.exists(output_directory):
            print(f'The directory {output_directory} already exists.')
        os.makedirs(output_directory, exist_ok=True)
        self.object_images_directory = output_directory  # Keep a record of the output directory
        print("Cropping...")
        
        for image_name in tqdm(self.object_detections.keys(), desc="Cropping Objects", unit="Image"):
            img_path = os.path.join(self.directory, image_name)
            if not os.path.exists(img_path):
                continue

            img = cv2.imread(img_path)
            detections = self.object_detections[image_name]
            W, H = detections['image_size']

            for n, contours in enumerate(detections['polygons']):
                instance_mask = np.zeros((W,H), dtype = np.uint8)
                cv2.drawContours(instance_mask, contours, contourIdx=-1, color=255, thickness=cv2.FILLED)
                instance_mask = cv2.resize(instance_mask,(img.shape[1],img.shape[0]))
                rescaled_contours, _ = cv2.findContours(instance_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
                x, y, w, h = cv2.boundingRect(rescaled_contours[0])
                x_min, x_max = x , x+w
                y_min, y_max = y , y+h
                cropped_img = img[int(y_min):int(y_max), int(x_min):int(x_max)]
           
                cv2.imwrite(os.path.join(output_directory, f"{os.path.splitext(image_name)[0]}_{object_name}_{n+1}.jpg"), cropped_img)
            
        return ImageDataset(output_directory)


    def export_train_data(self, output_dir: str, N: int, format: str = "COCO", include_flagged: bool = True, dataset_name: str = "exported_data"):
        if format not in ["COCO", "YOLO"]:
            raise ValueError("Only COCO and YOLO formats are supported.")

        # Prepare directories for output
        base_output_dir = os.path.join(output_dir, dataset_name)
        output_dir = base_output_dir
        n = 1
        while os.path.exists(output_dir):
            output_dir = f"{base_output_dir}_{n}"
            n += 1
        os.makedirs(output_dir, exist_ok=True)

        # Select images based on the include_flagged parameter
        flagged = self.flagged if include_flagged else []
        remaining = [img for img in self.images if img not in flagged]
        images_to_export = random.sample(flagged, min(N, len(flagged))) + random.sample(remaining, max(0, N - len(flagged)))

        if format == "COCO":
            coco_data = {
                "images": [],
                "annotations": [],
                "categories": [{"id": 1, "name": "object", "supercategory": "plant"}]
            }
            annotation_id = 1
            image_id = 1

            for image_name in tqdm(images_to_export, desc="Preparing COCO annotations", unit="image"):
                # Get image size
                img_path = os.path.join(self.directory, image_name)
                img = Image.open(img_path)
                width, height = (img.size[1], img.size[0]) if img._getexif().get(274) in [6, 8] else img.size

                # Add image information
                coco_data["images"].append({
                    "id": image_id,
                    "file_name": image_name,
                    "height": height,
                    "width": width
                })

                # Add annotations (bounding boxes)
                for box in self.object_detections.get(image_name, []):
                    x1, y1, x2, y2 = map(int, box)  # Get coordinates as integers
                    width_bbox = x2 - x1
                    height_bbox = y2 - y1

                    coco_data["annotations"].append({
                        "id": annotation_id,
                        "image_id": image_id,
                        "category_id": 1,  # 'leaf' category ID
                        "bbox": [x1, y1, width_bbox, height_bbox],  # COCO uses [x, y, width, height]
                        "area": width_bbox * height_bbox,  # Area of the bounding box
                        "iscrowd": 0  # 'iscrowd' is 0 for individual instances
                    })
                    annotation_id += 1

                image_id += 1

            # Save COCO JSON file
            with open(os.path.join(output_dir, "detections_coco.json"), 'w') as f:
                json.dump(coco_data, f, indent=4)

        elif format == "YOLO":
            for image_name in tqdm(images_to_export, desc="Preparing YOLO annotations", unit="image"):
                # Get image size
                img_path = os.path.join(self.directory, image_name)
                img = Image.open(img_path)
                width, height = (img.siz1e[1], img.size[0]) if img._getexif().get(274) in [6, 8] else img.size

                # YOLO annotation file
                yolo_file_path = os.path.join(output_dir, f"{os.path.splitext(image_name)[0]}.txt")
                with open(yolo_file_path, 'w') as yolo_file:
                    for box in self.detections.get(image_name, []):
                        x1, y1, x2, y2 = map(int, box)

                        # Convert bbox to YOLO format (class_id, x_center, y_center, bbox_width, bbox_height), normalized
                        x_center = (x1 + x2) / 2 / width
                        y_center = (y1 + y2) / 2 / height
                        bbox_width = (x2 - x1) / width
                        bbox_height = (y2 - y1) / height

                        # YOLO format: "class_id x_center y_center width height"
                        yolo_file.write(f"0 {x_center} {y_center} {bbox_width} {bbox_height}\n")  # Assuming 'leaf' has class ID 0

        # Copy referenced images with a progress bar
        for image_name in tqdm(images_to_export, desc="Copying images", unit="image"):
            shutil.copy(os.path.join(self.directory, image_name), output_dir)

    def object_dataset(self):
        objects_dir = self.object_images_directory
        object_dataset = ImageDataset(objects_dir)
        return object_dataset

    def segment(self, segmenter: Segmenter = None, use_saved = True, masks_file : str = None, background = False ):
       

        if use_saved:
            if not masks_file:
                masks_file =  os.path.join(self.directory,hidden_file_name, "masks.json")
            if os.path.exists(masks_file):
                self.read_segmentation(masks_file)
                return

        self.masks["channels"] = list(segmenter.channels)
        if background:
            self.masks["channels"].insert(0,"background")

        
        for image_name in tqdm(self.images, desc= 'Segmenting Images...',unit='image'):
           
            
            if not segmenter:
                print("No segmenter.")
                return
            
            image = cv2.imread(os.path.join(self.directory,image_name)) 
            output = segmenter.inference(image) 
            self.masks[image_name] = output
        
        self.write_segmentation()

    def write_segmentation(self, output_file : str = None, polys : bool = False):
        if not output_file:
            output_file = os.path.join(self.directory,hidden_file_name, 'masks.json')
        
        categories = []
        for n, category in enumerate(self.masks['channels']):
            cat = {"id" : n+1, "name":category, "supercategory": "mask"}
            categories.append(cat)

        today = datetime.date.today().isoformat()

        coco_data = {
            "info": {
                "description": "gwel-dataset",
                "version": "1.0",
                "year": datetime.date.today().year,
                "contributor": "",
                "date_created": today
            },
            "licenses": [],
            "images": [],
            "annotations": [],
            "categories": categories
        }

             
        annotation_id = 1
        image_id = 1

        for image_name in tqdm(self.images, desc="Writing COCO annotations", unit="image"):
            
            height, width = self.image_sizes[image_name]

            coco_data["images"].append({
                "id": image_id,
                "file_name": image_name,
                "height": height,
                "width": width
            })
            

            masks = self.masks.get(image_name, [])
            for n, channel in enumerate(self.masks['channels']):
                rle = masks[channel]
                rle['counts'] = rle['counts'].decode('utf-8')
                if rle['counts'] == 'Pfb\\1':
                    continue
               # if polys:   
                    #from shapely.geometry import Polygon
                #    binary_mask = maskUtils.decode(rle).astype(np.uint8)
                #    kernel = np.ones((3,3), np.uint8)
                 #   smoothed_mask = cv2.morphologyEx(binary_mask, cv2.MORPH_CLOSE, kernel, iterations=4)
                  #  cv2.imwrite(f'test-{n}-{image_id}.png', 255*smoothed_mask)
                    #contours, _ = cv2.findContours(binary_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
                    #simplified_polygons = []
                    #for contour in contours:
                     #   points = contour.squeeze()
                      #  if points.shape[0] < 3:
                       #     continue
                       # polygon = Polygon(points)
                       # simplified = polygon.simplify(tolerance=2.0, preserve_topology=True)
                       # simplified_coords = np.array(simplified.exterior.coords)
                       # simplified_polygons.append(simplified_coords) 
                    #polygons = []
                    #for contour in simplified_polygons:
                     #   contour = contour.reshape(-1, 2)
                      #  if len(contour) >= 3:  # Needs at least 3 points to form a polygon
                       #     polygon = contour.flatten().tolist()
                        #    polygons.append(polygon)
                    #segmentation = polygons
               # else:
                segmentation = rle
                
                coco_data["annotations"].append({
                    "id": annotation_id,
                    "image_id": image_id,
                    "category_id": n+1,  
                    "segmentation": segmentation, 
                    "bbox": maskUtils.toBbox(rle).tolist(),
                    "iscrowd": 1 
                })

                annotation_id += 1
        
            image_id += 1

        
        if not output_file:
            print('Could not write with no output file defined. For default write locations ensure overwrite = True')
            return

        with open(os.path.join(output_file), 'w') as f:
            json.dump(coco_data, f, indent=4)
	
    def read_segmentation(self, annotations_file: str = None):
        if annotations_file is None:
            annotations_file = os.path.join(self.directory, '.hidden', 'masks.json')

        with open(annotations_file, 'r') as f:
            coco_data = json.load(f)

        id_to_channel = {cat["id"]: cat["name"] for cat in coco_data["categories"]}

        image_info = {image["id"]: image["file_name"] for image in coco_data["images"]}
        self.masks = {'channels': list(id_to_channel.values())}

        for image_name in self.images:
            self.masks[image_name] = {}

        for ann in tqdm(coco_data["annotations"], desc="Loading annotations", unit="annotation"):
            image_id = ann["image_id"]
            channel = id_to_channel[ann["category_id"]]
            image_name = image_info[image_id]

            rle = ann["segmentation"]
            rle["counts"] = rle["counts"].encode("utf-8")  # required for pycocotools
            if image_name in self.images:
                self.masks[image_name][channel] = rle





import argparse
import os
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
import torch.nn as nn
from tqdm import tqdm
from src.UNET.unet import UNet
from src.UNET.data_loader import CocoSegmentationDataset
import numpy as np
import ast 
import pandas as pd



def iou_score(preds, targets, threshold=0.5):
    preds = torch.sigmoid(preds) > threshold
    targets = targets > threshold

    preds = preds.view(preds.size(0), preds.size(1), -1)
    targets = targets.view(targets.size(0), targets.size(1), -1)

    intersection = torch.sum(preds & targets, dim=2).float()
    union = torch.sum(preds | targets, dim=2).float()
    union = union * (union > 10)
    iou = intersection / union
    return np.array(torch.nanmean(iou,dim=0).cpu())

def train(model, train_loader, criterion, optimizer, device, epoch, iou_threshold = 0.5):
    model.train()
    running_loss = 0.0
    running_iou = np.array([0.0,0.0,0.0])
    total_samples = 0 
    total_samples_iou_score = np.array([0.0,0.0,0.0])
  
    for i, (inputs, targets) in enumerate(tqdm(train_loader)):
        inputs, targets = inputs.to(device), targets.to(device)
     
        optimizer.zero_grad()
        outputs = model(inputs)

        loss = criterion(outputs, targets)
        loss.backward()
        optimizer.step()
        
        iou = iou_score(outputs, targets, threshold=iou_threshold)
        running_loss += loss.item()

        running_iou = running_iou + np.nan_to_num(iou, nan=0.0)
        total_samples += 1
        total_samples_iou_score += (~np.isnan(iou)).astype(int)
        
    avg_loss = running_loss / total_samples
    avg_iou = running_iou / total_samples_iou_score
    
    print(f"Epoch {epoch}. Avg loss: {avg_loss:.3f}, Avg IoU 1: {avg_iou[0]:.3f}, Avg IoU 2: {avg_iou[1]:.3f}, Avg Iou 3: {avg_iou[2]:.3f}")
    
    df = pd.DataFrame([{
        'Epoch': epoch,
        'Avg Loss': avg_loss,
        'Avg IoU 1': avg_iou[0],
        'Avg IoU 2': avg_iou[1],
        'Avg IoU 3':avg_iou[2]
    }])

    return df

def validate(model, val_loader, criterion, device):
    model.eval()
    val_loss = 0.0
    with torch.no_grad():
        for inputs, targets in val_loader:
            inputs, targets = inputs.to(device), targets.to(device)
            outputs = model(inputs)
            loss = criterion(outputs, targets)
            val_loss += loss.item()

    print(f"Validation loss: {val_loss / len(val_loader):.4f}")

def main(args):
    
    categories = ast.literal_eval(args.categories)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    model = UNet(in_channels=args.in_channels, out_channels=len(categories)).to(device)

    criterion = nn.BCEWithLogitsLoss() #if args.out_channels == 1 else nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=args.lr)
    
    train_dataset = CocoSegmentationDataset(image_dir = args.train_images_dir, coco_json = args.train_annotations ,categories = categories)
   # val_dataset = CocoSegmentationDataset(image_dir = args.train_images_dir, coco_json = args.train_annotations ,category = args.category)

    train_loader = DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True)
    #val_loader = DataLoader(val_dataset, batch_size=args.batch_size, shuffle=False)

    # Training loop
    train_data = pd.DataFrame()
    for epoch in range(args.epochs):
        new_row = train(model, train_loader, criterion, optimizer, device, epoch)
        train_data = pd.concat([train_data, new_row], ignore_index=True)
        train_data.to_csv(args.train_metrics_path,index=False)
        # Validate at the end of each epoch
     #   if (epoch + 1) % args.val_interval == 0:
      #      print("Validating...")
       #     validate(model, val_loader, criterion, device)

        # Save the model checkpoint after each epoch
        if (epoch + 1) % args.save_interval == 0:
            checkpoint_path = os.path.join(args.save_dir, f"unet_epoch_{epoch + 1}.pth")
            torch.save(model.state_dict(), checkpoint_path)
            print(f"Saved checkpoint: {checkpoint_path}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Train a U-Net model")

    parser.add_argument('-td','--train_images_dir', type=str, required=True, help="Path to directory of training images")
    parser.add_argument('-ta','--train_annotations', type=str, required=True, help="Path to training annotations")
    parser.add_argument('-vd','--val_images_dir', type=str, required=True, help="Path to directory of validation images")
    parser.add_argument('-va','--val_annotations', type=str, required=True, help="Path to validation annotations")
    parser.add_argument('-c','--categories', type=str, required=True, help="Integer tag for annotation categories")
    parser.add_argument('-s','--save_dir', type=str, required=True, help="Directory to save model checkpoints")

    parser.add_argument('--epochs', type=int, default=10, help="Number of epochs")
    parser.add_argument('--batch_size', type=int, default=16, help="Batch size")
    parser.add_argument('--lr', type=float, default=1e-4, help="Learning rate")
    parser.add_argument('--in_channels', type=int, default=1, help="Number of input channels (1 for grayscale images)")
    parser.add_argument('--out_channels', type=int, default=1, help="Number of output channels (1 for binary segmentation)")
    parser.add_argument('--save_interval', type=int, default=1, help="How often to save model checkpoints (in epochs)")
    parser.add_argument('--val_interval', type=int, default=1, help="How often to validate the model (in epochs)")
    parser.add_argument('--train_metrics_path',type=str,default="train-data.csv",help="Path to output train metrics csv")
    args = parser.parse_args()

    # Check that paths exist
    assert os.path.exists(args.train_images_dir), f"Train images path '{args.train_images_dir}' not found."
    assert os.path.exists(args.train_annotations), f"Train masks path '{args.train_annotations}' not found."
    assert os.path.exists(args.val_images_dir), f"Validation images path '{args.val_images_dir}' not found."
    assert os.path.exists(args.val_annotations), f"Validation masks path '{args.val_annotations}' not found."
    os.makedirs(args.save_dir, exist_ok=True)

    # Start training
    main(args)




from typing import Literal
import random
import copy
import cv2
import os
import numpy as np
import pandas as pd
from PIL import Image
from tqdm import tqdm 
from gwel.dataset import ImageDataset 
import numpy as np
import pycocotools.mask as mask_utils

class Viewer:
    def __init__(self, dataset: ImageDataset,
                 mode: Literal['', 'instance', 'segmentation','coordinates'] = ''  ,
                 max_pixels: int = 800,
                 contour_thickness: int = 1,
                 col_scheme = {}):
        
        self.max_pixels = max_pixels 
        self.contour_thickness = contour_thickness
        self.col_scheme = col_scheme
        self.mode = mode
        self.dataset = dataset
        self.directory = getattr(self.dataset, "resized_images_directory", self.dataset.directory)
        self.images = self.dataset.images
        self.total_images = len(self.images)
        self.index = 0 
        self.col = False
        
    def load_image(self):
        self.image_name = self.images[self.index]
        self.image_path = os.path.join(self.directory,self.image_name)
        self.image = cv2.imread(self.image_path)
        self.image_size = self.dataset.image_sizes[self.image_name]
        self.flagged = self.image_name in self.dataset.flagged 
        if self.image is None:
            print(f"Warning: Unable to read image {self.image_name}.")
            return False
            
    def open(self):

        while True:
            self.load_image() 
            self.style() 
            self.display_image()
           
            key = cv2.waitKey(0)
            if key == ord('q'):  # Quit the viewer
                break
            elif key == ord('p'):  # previous image
                self.index = max(self.index - 1,0)
            elif key == ord('n'):  # next image
                self.index = min (self.index + 1 , self.total_images-1)  
            elif key == ord('f'):  # toggle flag/unflag the image
                if self.image_name in self.dataset.flagged:
                    self.dataset.flagged.remove(self.image_name)
                else:
                    self.dataset.flagged.append(self.image_name)
            elif key == ord('c'):
                self.col = not self.col    # toggle between colour schemes
            elif key == ord('1'):
                self.mode = "instance"
            elif key == ord('2'):
                self.mode = "segmentation"
            elif key == ord('3'):
                self.mode = "seginstance"
            elif key == ord("4"):
                self.mode = "instandseg"
            elif key == ord("5"):
                self.mode = "circandseg"
            elif key == ord("0"):
                self.mode = ""
            


        cv2.destroyAllWindows()


    def display_image(self):
        
        height, width = self.image.shape[:2]
        scale_ratio = min(self.max_pixels / width, self.max_pixels / height, 1)
        scaled_image = cv2.resize(self.image, (int(width * scale_ratio), int(height * scale_ratio)), interpolation=cv2.INTER_LINEAR)

        bar_height = 40
        bar_color = (142, 76, 195) if self.flagged else (60, 179, 113)
        bar = np.full((bar_height, scaled_image.shape[1], 3), bar_color, dtype=np.uint8)
        overlay_text = f"{self.image_name} | {self.index + 1}/{self.total_images}"
        text_color = (255, 255, 255)
        cv2.putText(bar, overlay_text, (10, 30), cv2.FONT_HERSHEY_SIMPLEX, 0.7, text_color, 2)
        combined_image = cv2.vconcat([bar, scaled_image])

        cv2.imshow("Viewer", combined_image)

        return True

    def style(self):
        
        if self.mode == "instance": 
            self.detections = copy.deepcopy(self.dataset.object_detections[self.image_name])
            
            if self.detections['image_size']:     
                W, H = self.detections['image_size']
                colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                for contours in self.detections['polygons']:
                    if self.col:
                        colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                    instance_mask = np.zeros((W,H), dtype = np.uint8)
                    cv2.drawContours(instance_mask, contours, contourIdx=-1, color=255, thickness=cv2.FILLED)
                    instance_mask = cv2.resize(instance_mask,(self.image.shape[1],self.image.shape[0]))
                    rescaled_contours, _ = cv2.findContours(instance_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
                    cv2.drawContours(self.image, rescaled_contours, -1, colour , self.contour_thickness)
        
        if self.mode == "segmentation":
            rles_dict = self.dataset.masks[self.image_name] 
            rles = list(rles_dict.values())
            for rle in rles:
                mask = mask_utils.decode(rle) 
                mask = cv2.resize(mask,(self.image.shape[1],self.image.shape[0]))
                contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
                colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                cv2.drawContours(self.image,contours,-1, colour , self.contour_thickness)
       
        

        if self.mode == "seginstance":
            rles_dict = self.dataset.masks[self.image_name] 
            rles = list(rles_dict.values())
             
            instance_mask = np.ones(self.image.shape[:2], dtype=np.uint8)
             
            if self.image_name in self.dataset.object_detections:
                self.detections = copy.deepcopy(self.dataset.object_detections[self.image_name])
                W, H = self.detections['image_size']
                instance_mask = np.zeros((W,H), dtype = np.uint8)
                for contours in self.detections['polygons']:
                    cv2.drawContours(instance_mask, contours, contourIdx=-1, color=255, thickness=cv2.FILLED)
                instance_mask = cv2.resize(instance_mask,(self.image.shape[1],self.image.shape[0]))
            
            for rle in rles:
                mask = mask_utils.decode(rle) 
                mask = cv2.resize(mask,(self.image.shape[1],self.image.shape[0]))
                mask = mask * instance_mask
                contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
                colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                cv2.drawContours(self.image,contours,-1, colour , self.contour_thickness)
            
 
        if self.mode == "instandseg": 
            self.detections = copy.deepcopy(self.dataset.object_detections[self.image_name])
            
            if self.detections['image_size']:     
                W, H = self.detections['image_size']
                colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                for contours in self.detections['polygons']:
                    if self.col:
                        colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                    instance_mask = np.zeros((W,H), dtype = np.uint8)
                    cv2.drawContours(instance_mask, contours, contourIdx=-1, color=255, thickness=cv2.FILLED)
                    instance_mask = cv2.resize(instance_mask,(self.image.shape[1],self.image.shape[0]))
                    rescaled_contours, _ = cv2.findContours(instance_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
                    cv2.drawContours(self.image, rescaled_contours, -1, colour , self.contour_thickness)
        
            rles_dict = self.dataset.masks[self.image_name] 
            rles = list(rles_dict.values())
             
            for rle in rles:
                mask = mask_utils.decode(rle) 
                mask = cv2.resize(mask,(self.image.shape[1],self.image.shape[0]))
                contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
                colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                cv2.drawContours(self.image,contours,-1, colour , self.contour_thickness)
       
        if self.mode == "circandseg": 
            self.detections = copy.deepcopy(self.dataset.object_detections[self.image_name])
            
            if self.detections['image_size']:     
                W, H = self.detections['image_size']
                colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                for contours in self.detections['polygons']:
                    if self.col:
                        colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                    instance_mask = np.zeros((W,H), dtype = np.uint8)
                    cv2.drawContours(instance_mask, contours, contourIdx=-1, color=255, thickness=cv2.FILLED)
                    instance_mask = cv2.resize(instance_mask,(self.image.shape[1],self.image.shape[0]))
                    rescaled_contours, _ = cv2.findContours(instance_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
                    x, y, w, h = cv2.boundingRect(rescaled_contours[0])
                    center = (x + w // 2, y + h // 2)
                    axes = (w // 2, h // 2)
                    cv2.ellipse(self.image, center, axes, 0, 0, 360, colour, self.contour_thickness)
                    #cv2.drawContours(self.image, rescaled_contours, -1, colour , self.contour_thickness)
        
            rles_dict = self.dataset.masks[self.image_name] 
            rles = list(rles_dict.values())
             
            for rle in rles:
                mask = mask_utils.decode(rle) 
                mask = cv2.resize(mask,(self.image.shape[1],self.image.shape[0]))
                contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
                colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
                cv2.drawContours(self.image,contours,-1, colour , self.contour_thickness)

            #height, width = self.image.shape[:2]
            #scale_ratio = min(800 / width, 800 / height, 1)
            #self.image = cv2.resize(self.image, (int(width * scale_ratio), int(height * scale_ratio)), interpolation=cv2.INTER_LINEAR)

           #self.image = cv2.resize(self.image,self.dataset.image_sizes[self.image_name])

            
    def save(self, output_path):
        self.load_image() 
        self.style() 
        cv2.imwrite(output_path, self.image)



"""
    def _draw_contours(self,image,contour):
        image = np.array(image)
        masks = cv2.imread(mask_path).transpose(2,0,1)
        if masks is None:
            print(f"Warning: Unable to read mask for {image_name}.")
            return False

        masks_split = np.split(masks,masks.shape[0])
        colours = [(0,255,255),(255,255,0),(255,0,255)]

        image_with_contour = image.copy()

        for idx in range(masks.shape[0]):
            mask = masks_split[idx].astype(np.uint8).squeeze()
            contours, _  = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
            cv2.drawContours(image_with_contour, contours, -1, colours[idx], 5)
        image_with_contour = Image.fromarray(image_with_contour)
        return image_with_contour


    def resize_width_keep_aspect(self,img, new_width):
        orig_width, orig_height = img.size
        new_height = int((new_width / orig_width) * orig_height)
        return img.resize((new_width, new_height), Image.LANCZOS)


    def _centre_pad(self,image,size):
        width, height = image.size
        image = np.array(image)
        target_height = size[0]
        target_width = size[1]
        top = (target_height - height) // 2
        bottom = target_height - height - top
        left = (target_width - width) // 2
        right = target_width - width - left
        padded_image = cv2.copyMakeBorder(
            image, top, bottom, left, right, 
            borderType=cv2.BORDER_CONSTANT, 
            value=(255, 255, 255)        )
        padded_image = Image.fromarray(padded_image)
        return padded_image

    def extend_image_with_number(self,image, number, font_scale=7, thickness=10):
        # Convert PIL image to OpenCV format if it's a PIL image
        if isinstance(image, Image.Image):
            image = np.array(image)
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)  # Convert RGB (PIL) to BGR (OpenCV)
        
        h, w, c = image.shape
        new_w = int(w * 1.1)  # Extend by 10%
        
        # Create a new white image
        extended_image = np.ones((h, new_w, c), dtype=np.uint8) * 255
        extended_image[:, new_w-w:] = image  # Copy the original image
        
        # Put the number in the center of the new area
        text = str(number)
        font = cv2.FONT_HERSHEY_SIMPLEX
        text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]
        text_x = (new_w - w - text_size[0]) // 2
        text_y = (h + text_size[1]) // 2
        
        cv2.putText(extended_image, text, (text_x, text_y), font, font_scale, (0, 0, 0), thickness)

        # Convert back to PIL format
        pil_image = Image.fromarray(cv2.cvtColor(extended_image, cv2.COLOR_BGR2RGB))  # Convert BGR to RGB for PIL
        return pil_image
       
    def animate(self,output_dir, directory, image_list, image_sizes, factors,masks,label = True):
        os.makedirs(os.path.join(output_dir),exist_ok=True)
        df = pd.DataFrame(factors)
        df['image_name']=image_list
        factors_list = list(factors.keys())
        if 'time' not in factors_list:
            print('Time is not a factor. Please add "time" as a factor')
            return
        else:
            factors_list.remove('time')
                  
        for group_name, group_df in tqdm(df.groupby(factors_list)):
            images = list(group_df['image_name'])
            time = list(group_df['time'])
            images = [image for _, image in sorted(zip(time,images))]
            time =  [t for t, _ in sorted(zip(time,images))]


            # get the largest size (use padding)
            widths = []
            heights = []
            for image in images:
                size = image_sizes.get(image)
                widths.append(size[0])
                heights.append(size[1])
            
            width = int(np.max(widths))

            height = int(np.max(np.array(heights)/np.array(widths)*width))


            first_image = Image.open(os.path.join(directory,images[0]))
            if masks:
                first_image = self._draw_contours(first_image,masks.get(images[0]))
                first_image = self.resize_width_keep_aspect(first_image,width)
                first_image = self._centre_pad(first_image,(height,width))
                if label:
                    first_image = self.extend_image_with_number(first_image,time[0])
            gif_images = [first_image]

            for i, image in enumerate(images[1:]):
                img = Image.open(os.path.join(directory,image))
                if masks:
                    img = self._draw_contours(img,masks.get(image))
                img  = self.resize_width_keep_aspect(img ,width)
                img = self._centre_pad(img,(height,width)) 
                img = self.extend_image_with_number(img,time[i+1])
                gif_images.append(img)
            
           
            group = "_".join(f"{f}_{v}" for f, v in zip(factors_list, group_name))  
            times = "_".join(f"{t}" for t in time)  
            gif_filename =os.path.join(output_dir,f'{group}'+'_time_'+f'{times}.gif')
        
            gif_images[0].save(
               gif_filename,
                save_all=True,
                append_images=gif_images[1:],  # Append the rest of the images
                duration=500,  # Duration of each frame in milliseconds
                loop=0  # Infinite loop (set to a specific number for a limited number of loops)
            )

        print("GIF created successfully!")
            # group by non-time factors
            #loop over them and add image to animation
            # if masks specfied before add image add contours
            # save animation to output_Dir
                


"""

       
